package org.robokind.demo.robot.replication;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;
import org.jflux.api.core.config.Configuration;
import org.jflux.avrogen.DoubleRecord;
import org.jflux.avrogen.EventListRecord;
import org.jflux.avrogen.EventRecord;
import org.jflux.avrogen.HeaderRecord;
import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.robokind.api.common.lifecycle.config.RKManagedGroupConfigUtils;
import org.robokind.api.common.lifecycle.utils.ManagedServiceFactory;
import org.robokind.api.common.lifecycle.utils.ManagedServiceGroup;
import org.robokind.api.common.osgi.lifecycle.OSGiComponentFactory;
import org.robokind.api.common.position.DoubleRange;
import org.robokind.api.common.position.NormalizableRange;
import org.robokind.api.common.position.NormalizableRange.DefaultRange;
import org.robokind.api.common.position.NormalizedDouble;
import org.robokind.api.common.utils.Utils;
import org.robokind.api.motion.AbstractJoint;
import org.robokind.api.motion.Joint;
import org.robokind.api.motion.messaging.RemoteJointProperty;
import org.robokind.demo.robot.replication.messaging.PortableNormalizedDouble;
import org.robokind.impl.messaging.JMSAvroServiceFacade;
import org.robokind.impl.messaging.config.RKMessagingConfigUtils;

public class Activator implements BundleActivator {
    private final static String theGroupId = "robotReplicationDemoGroup";

    @Override
    public void start(BundleContext bc) throws Exception {
        String destConfigId = RobotReplicator.DESTINATION_CONFIG_ID;
        ManagedServiceFactory fact = new OSGiComponentFactory(bc);
        registerEncodingConfigs(fact);
        
        RKMessagingConfigUtils.registerConnectionConfig(
                RobotReplicator.REPLICATION_CONNECTION_CONFIG_ID, 
                "127.0.0.1", null, fact);
        RKMessagingConfigUtils.registerTopicConfig(
                destConfigId, "robotReplication", null, fact);
        replicate(bc, fact, RobotReplicator.REPLICATION_CONNECTION_CONFIG_ID, 
                RobotReplicator.RECEIVER_CONNECTION_CONFIG_ID, destConfigId,
                200L);
        
        //test(bc, connectConfigId, destConfigId, fact);
    }
    
    private void replicate(BundleContext context, ManagedServiceFactory fact,
            String replicationConnectConfigId, String receiverConnectConfigId,
            String destConfigId, long replicationPeriod){
        List<Configuration<String>> configs = 
                new ArrayList<Configuration<String>>();
        configs.add(RobotReplicator.buildRobotReceiverLifecycleConfig(
                context, receiverConnectConfigId, destConfigId));
        configs.add(RobotReplicator.buildRobotReplicationLifecycleConfig(
                context, replicationConnectConfigId, destConfigId, replicationPeriod));
        
        Configuration<String> groupConf = 
                RKManagedGroupConfigUtils.buildManagedGroupConfig(
                        theGroupId, null, configs);
        ManagedServiceGroup group = 
                RKManagedGroupConfigUtils.buildGroup(fact, groupConf);
        group.start();
    }
    
    private void test(BundleContext bc, ManagedServiceFactory fact,
            String connectConfigId, String destConfigId){
        Joint j1a = new RJoint1(0);
        Joint j1b = new RJoint1(1);
        Joint j1c = new RJoint1(2);
        Joint j2a = new RJoint2(0);
        Joint j2b = new RJoint2(1);
        Joint j2c = new RJoint2(2);        
        RobotReplicator.startRobotReplicator(
                bc, theGroupId, connectConfigId, destConfigId, 
                Arrays.asList(j1a,j1b,j1c), 100L, null, fact);
        
        RobotReplicator.startRobotReceiver(
                theGroupId, connectConfigId, destConfigId, 
                Arrays.asList(j2a,j2b,j2c), null, fact);
    }
    
    @Override
    public void stop(BundleContext context) throws Exception {
    }
    
    private void registerEncodingConfigs(ManagedServiceFactory fact){
        RKMessagingConfigUtils.registerAvroSerializationConfig(
                EventListRecord.class, 
                EventListRecord.class, 
                EventListRecord.SCHEMA$, 
                null, null, 
                "application/avro-event-list", null, fact);
        RKMessagingConfigUtils.registerAvroSerializationConfig(
                EventRecord.class, 
                EventRecord.class, 
                EventRecord.SCHEMA$, 
                null, null, 
                "application/avro-event", null, fact);
        RKMessagingConfigUtils.registerAvroSerializationConfig(
                HeaderRecord.class, 
                HeaderRecord.class, 
                HeaderRecord.SCHEMA$, 
                null, null, 
                "application/avro-header", null, fact);
        RKMessagingConfigUtils.registerAvroSerializationConfig(
                NormalizedDouble.class, 
                DoubleRecord.class, 
                DoubleRecord.SCHEMA$,
                new PortableNormalizedDouble.MessageRecordAdapter(), 
                new PortableNormalizedDouble.RecordMessageAdapter(), 
                JMSAvroServiceFacade.AVRO_MIME_TYPE, null, fact);
    }
    
    static class RJoint extends AbstractJoint {
        
        public RJoint(int id){
            super(new Joint.Id(id));
            this.addProperty(new RemoteJointProperty("test", "test", NormalizedDouble.class, new NormalizedDouble(0.2), new DefaultRange()));
        }

        @Override
        public String getName() {
            return "name";
        }

        @Override
        public NormalizedDouble getDefaultPosition() {
            return new NormalizedDouble(0.5);
        }

        @Override
        public NormalizableRange<Double> getPositionRange() {
            return new DoubleRange(0, 1);
        }

        @Override
        public NormalizedDouble getGoalPosition() {
            return new NormalizedDouble(0.2);
        }

        @Override
        public void setEnabled(Boolean enabled) {
        }

        @Override
        public Boolean getEnabled() {
            return true;
        }
    }
    
    static class RJoint1 extends RJoint {
        public RJoint1(final int id){
            super(id);
            final Random rand = new Random();
            this.addProperty(new RemoteJointProperty("currentPosition", "currentPosition", NormalizedDouble.class, new NormalizedDouble(0.0), new DefaultRange()){

                @Override
                public NormalizedDouble getValue() {
                    double x = rand.nextDouble();
                    double j = ((double)id)/10.0;
                    x /= 100.0;
                    x += j;
                    return new NormalizedDouble(Utils.bound(x, 0.0, 1.0));
                }
                
            });
        }
    }
    
    static class RJoint2 extends RJoint {
        public RJoint2(final int id){
            super(id);
            this.addProperty(new RemoteJointProperty("currentPosition", "currentPosition", NormalizedDouble.class, new NormalizedDouble(0.0), new DefaultRange()));
        }
    }
    
//    @Override
//    public void start(BundleContext context) throws Exception {
//        ManagedServiceFactory fact = new OSGiComponentFactory(context);
//        initialize(context, fact);
//        
//        String commandBufferId = "commandBufferId";
//        Class<ServiceCommand> commandClass = ServiceCommand.class;
//        
//        ReceiverTest.createRoute(
//                commandBufferId, commandClass.toString(), 
//                new ReceiverTest.TestEventLogger(), fact);  
//        
//        createTestBuffer(commandBufferId, 
//                25, 200, commandClass, 
//                new Source<ServiceCommand>() {
//                    private long count = 0;
//                    @Override public ServiceCommand getValue() {
//                        return new PortableServiceCommand(
//                                theGroupId, "dest", 
//                                System.currentTimeMillis(), "command " + count++);
//                    }}, fact);
//        
//        String errorBufferId = "errorBufferId";
//        Class<ServiceError> errorClass = ServiceError.class;
//        
//        ReceiverTest.createRoute(
//                errorBufferId, errorClass.toString(), 
//                new ReceiverTest.TestErrorLogger(), fact);
//        
//        createTestBuffer(errorBufferId, 
//                15, 700, errorClass, 
//                new Source<ServiceError>() {
//                    private long count = 0;
//                    @Override public ServiceError getValue() {
//                        return new PortableServiceError(
//                                theGroupId, "dest", System.currentTimeMillis(), 
//                                "error " + count++, "This is just a test.",
//                                "Everything is fine, please remain calm.");
//                    }}, fact);  
//    }
//    
//    private void initialize(BundleContext context, ManagedServiceFactory fact){
//        registerEncodingConfigs(fact);
//        startServicePanel(context);
//        
//        RKMessagingConfigUtils.registerConnectionConfig(
//                "testConnectConfig", "127.0.0.1", null, fact);
//        
//        startReplicator(context, fact);
//        ReceiverTest.start(fact);        
//    }
//    
//    private <T> void createTestBuffer(
//            String bufferId, int bufferCapacity, long timerInterval, 
//            Class<T> messageClass, Source<T> testSource, 
//            ManagedServiceFactory fact){
//        List<Configuration<String>> configs = new ArrayList(
//                ReplicationSenderGroupLifecycleUtils.getReplicationBufferLifecycleConfigs(
//                        theGroupId, bufferId, 
//                        messageClass.toString(), null, bufferCapacity));
//        configs.add(NodeLinkLifecycleUtils.producerConsumerLinkLifecycleConfig(
//                bufferId + "LinkId", null, 
//                bufferId + "TimerId", bufferId + "TimerId", null, 
//                EventBufferUtils.PROP_EVENT_BUFFER_NODE_ID, bufferId, null));
//        configs.add(CollectorLifecycleUtils.buildIntervalTimerLifecycleConfig(
//                bufferId + "TimerId", bufferId + "TimerId", 
//                1000, timerInterval, testSource, null));
//        Configuration<String> groupConf = 
//                RKManagedGroupConfigUtils.buildManagedGroupConfig(
//                        theGroupId, null, configs);
//        ManagedServiceGroup group = 
//                RKManagedGroupConfigUtils.buildGroup(fact, groupConf);
//        group.start();   
//    }
//    
//    private void startReplicator(BundleContext context, ManagedServiceFactory fact){
//        RKMessagingConfigUtils.registerQueueConfig("testDest", "test", null, fact);
//        List<Configuration<String>> lifecycles = 
//                getLifecycleConfigs(context, theGroupId, "testDest", "testConnectConfig");
//        Configuration<String> groupConf = 
//                RKManagedGroupConfigUtils.buildManagedGroupConfig(
//                        theGroupId, null, lifecycles);
//        ManagedServiceGroup group = RKManagedGroupConfigUtils.buildGroup(fact, groupConf);
//        group.start();        
//    }
//    
//    private List<Configuration<String>> getLifecycleConfigs(
//            BundleContext context, String groupId, 
//            String destConfigId, String connectionConfigId){
//        return new ListMergeAdapter().adapt(Arrays.asList(
//                ReplicationSenderGroupLifecycleUtils.getJMSLifecycleConfigs(
//                        groupId, destConfigId, connectionConfigId),
//                ReplicationSenderGroupLifecycleUtils.getReplicationSenderLifecycleConfigs(
//                        groupId),
//                ReplicationSenderGroupLifecycleUtils.getCollectorLifecycleConfigs(
//                        groupId, context, null, 1000L, 2000L, null)
//            ));
//    }
//
//    private void startServicePanel(final BundleContext context) {
//        try{
//            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
//        }catch(Exception ex){}
//        
//        java.awt.EventQueue.invokeLater(new Runnable() {
//
//            @Override
//            public void run() {
//                ServicesFrame sf = new ServicesFrame();
//                sf.setBundleContext(context);
//                sf.setVisible(true);
//            }
//        });
//    }
}
